# Copyright (C) 2018-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest
import json
import msgpack

from flask import url_for

from swh.core.api import remote_api_endpoint, RPCServerApp
from swh.core.api import negotiate, JSONFormatter, MsgpackFormatter
from .test_serializers import ExtraType, extra_encoders, extra_decoders


class MyRPCServerApp(RPCServerApp):
    extra_type_encoders = extra_encoders
    extra_type_decoders = extra_decoders


@pytest.fixture
def app():
    class TestStorage:
        @remote_api_endpoint('test_endpoint_url')
        def test_endpoint(self, test_data, db=None, cur=None):
            assert test_data == 'spam'
            return 'egg'

        @remote_api_endpoint('path/to/endpoint')
        def something(self, data, db=None, cur=None):
            return data

        @remote_api_endpoint('serializer_test')
        def serializer_test(self, data, db=None, cur=None):
            assert data == ['foo', ExtraType('bar', b'baz')]
            return ExtraType({'spam': 'egg'}, 'qux')

    return MyRPCServerApp('testapp', backend_class=TestStorage)


def test_api_endpoint(flask_app_client):
    res = flask_app_client.post(
        url_for('something'),
        headers=[('Content-Type', 'application/json'),
                 ('Accept', 'application/json')],
        data=json.dumps({'data': 'toto'}),
    )
    assert res.status_code == 200
    assert res.mimetype == 'application/json'


def test_api_nego_default(flask_app_client):
    res = flask_app_client.post(
        url_for('something'),
        headers=[('Content-Type', 'application/json')],
        data=json.dumps({'data': 'toto'}),
    )
    assert res.status_code == 200
    assert res.mimetype == 'application/json'
    assert res.data == b'"toto"'


def test_api_nego_accept(flask_app_client):
    res = flask_app_client.post(
        url_for('something'),
        headers=[('Accept', 'application/x-msgpack'),
                 ('Content-Type', 'application/x-msgpack')],
        data=msgpack.dumps({'data': 'toto'}),
    )
    assert res.status_code == 200
    assert res.mimetype == 'application/x-msgpack'
    assert res.data == b'\xa4toto'


def test_rpc_server(flask_app_client):
    res = flask_app_client.post(
        url_for('test_endpoint'),
        headers=[('Content-Type', 'application/x-msgpack'),
                 ('Accept', 'application/x-msgpack')],
        data=b'\x81\xa9test_data\xa4spam')

    assert res.status_code == 200
    assert res.mimetype == 'application/x-msgpack'
    assert res.data == b'\xa3egg'


def test_rpc_server_extra_serializers(flask_app_client):
    res = flask_app_client.post(
        url_for('serializer_test'),
        headers=[('Content-Type', 'application/x-msgpack'),
                 ('Accept', 'application/x-msgpack')],
        data=b'\x81\xa4data\x92\xa3foo\x82\xc4\x07swhtype\xa9extratype'
             b'\xc4\x01d\x92\xa3bar\xc4\x03baz')

    assert res.status_code == 200
    assert res.mimetype == 'application/x-msgpack'
    assert res.data == (
        b'\x82\xc4\x07swhtype\xa9extratype\xc4'
        b'\x01d\x92\x81\xa4spam\xa3egg\xa3qux')


def test_api_negotiate_no_extra_encoders(app, flask_app_client):
    url = '/test/negotiate/no/extra/encoders'

    @app.route(url, methods=['POST'])
    @negotiate(MsgpackFormatter)
    @negotiate(JSONFormatter)
    def endpoint():
        return 'test'

    res = flask_app_client.post(
        url,
        headers=[('Content-Type', 'application/json')],
    )
    assert res.status_code == 200
    assert res.mimetype == 'application/json'
    assert res.data == b'"test"'



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import inspect
import os.path
import tempfile
import unittest
from unittest.mock import Mock, MagicMock

from hypothesis import strategies, given
import psycopg2
import pytest

from swh.core.db import BaseDb
from swh.core.db.common import db_transaction, db_transaction_generator
from .db_testing import (
    SingleDbTestFixture, db_create, db_destroy, db_close,
)


INIT_SQL = '''
create table test_table
(
    i       int,
    txt     text,
    bytes   bytea
);
'''

db_rows = strategies.lists(strategies.tuples(
    strategies.integers(-2147483648, +2147483647),
    strategies.text(
        alphabet=strategies.characters(
            blacklist_categories=['Cs'],  # surrogates
            blacklist_characters=[
                '\x00',  # pgsql does not support the null codepoint
                '\r',  # pgsql normalizes those
            ]
        ),
    ),
    strategies.binary(),
))


@pytest.mark.db
def test_connect():
    db_name = db_create('test-db2', dumps=[])
    try:
        db = BaseDb.connect('dbname=%s' % db_name)
        with db.cursor() as cur:
            cur.execute(INIT_SQL)
            cur.execute("insert into test_table values (1, %s, %s);",
                        ('foo', b'bar'))
            cur.execute("select * from test_table;")
            assert list(cur) == [(1, 'foo', b'bar')]
    finally:
        db_close(db.conn)
        db_destroy(db_name)


@pytest.mark.db
class TestDb(SingleDbTestFixture, unittest.TestCase):
    TEST_DB_NAME = 'test-db'

    @classmethod
    def setUpClass(cls):
        with tempfile.TemporaryDirectory() as td:
            with open(os.path.join(td, 'init.sql'), 'a') as fd:
                fd.write(INIT_SQL)

            cls.TEST_DB_DUMP = os.path.join(td, '*.sql')

            super().setUpClass()

    def setUp(self):
        super().setUp()
        self.db = BaseDb(self.conn)

    def test_initialized(self):
        cur = self.db.cursor()
        cur.execute("insert into test_table values (1, %s, %s);",
                    ('foo', b'bar'))
        cur.execute("select * from test_table;")
        self.assertEqual(list(cur), [(1, 'foo', b'bar')])

    def test_reset_tables(self):
        cur = self.db.cursor()
        cur.execute("insert into test_table values (1, %s, %s);",
                    ('foo', b'bar'))
        self.reset_db_tables('test-db')
        cur.execute("select * from test_table;")
        self.assertEqual(list(cur), [])

    @given(db_rows)
    def test_copy_to(self, data):
        # the table is not reset between runs by hypothesis
        self.reset_db_tables('test-db')

        items = [dict(zip(['i', 'txt', 'bytes'], item)) for item in data]
        self.db.copy_to(items, 'test_table', ['i', 'txt', 'bytes'])

        cur = self.db.cursor()
        cur.execute('select * from test_table;')
        self.assertCountEqual(list(cur), data)

    def test_copy_to_thread_exception(self):
        data = [(2**65, 'foo', b'bar')]

        items = [dict(zip(['i', 'txt', 'bytes'], item)) for item in data]
        with self.assertRaises(psycopg2.errors.NumericValueOutOfRange):
            self.db.copy_to(items, 'test_table', ['i', 'txt', 'bytes'])


def test_db_transaction(mocker):
    expected_cur = object()

    called = False

    class Storage:
        @db_transaction()
        def endpoint(self, cur=None, db=None):
            nonlocal called
            called = True
            assert cur is expected_cur

    storage = Storage()

    # 'with storage.get_db().transaction() as cur:' should cause
    # 'cur' to be 'expected_cur'
    db_mock = Mock()
    db_mock.transaction.return_value = MagicMock()
    db_mock.transaction.return_value.__enter__.return_value = expected_cur
    mocker.patch.object(
        storage, 'get_db', return_value=db_mock, create=True)

    put_db_mock = mocker.patch.object(
        storage, 'put_db', create=True)

    storage.endpoint()

    assert called
    put_db_mock.assert_called_once_with(db_mock)


def test_db_transaction__with_generator():
    with pytest.raises(ValueError, match='generator'):
        class Storage:
            @db_transaction()
            def endpoint(self, cur=None, db=None):
                yield None


def test_db_transaction_signature():
    """Checks db_transaction removes the 'cur' and 'db' arguments."""
    def f(self, foo, *, bar=None):
        pass
    expected_sig = inspect.signature(f)

    @db_transaction()
    def g(self, foo, *, bar=None, db=None, cur=None):
        pass

    actual_sig = inspect.signature(g)

    assert actual_sig == expected_sig


def test_db_transaction_generator(mocker):
    expected_cur = object()

    called = False

    class Storage:
        @db_transaction_generator()
        def endpoint(self, cur=None, db=None):
            nonlocal called
            called = True
            assert cur is expected_cur
            yield None

    storage = Storage()

    # 'with storage.get_db().transaction() as cur:' should cause
    # 'cur' to be 'expected_cur'
    db_mock = Mock()
    db_mock.transaction.return_value = MagicMock()
    db_mock.transaction.return_value.__enter__.return_value = expected_cur
    mocker.patch.object(
        storage, 'get_db', return_value=db_mock, create=True)

    put_db_mock = mocker.patch.object(
        storage, 'put_db', create=True)

    list(storage.endpoint())

    assert called
    put_db_mock.assert_called_once_with(db_mock)


def test_db_transaction_generator__with_nongenerator():
    with pytest.raises(ValueError, match='generator'):
        class Storage:
            @db_transaction_generator()
            def endpoint(self, cur=None, db=None):
                pass


def test_db_transaction_generator_signature():
    """Checks db_transaction removes the 'cur' and 'db' arguments."""
    def f(self, foo, *, bar=None):
        pass
    expected_sig = inspect.signature(f)

    @db_transaction_generator()
    def g(self, foo, *, bar=None, db=None, cur=None):
        yield None

    actual_sig = inspect.signature(g)

    assert actual_sig == expected_sig



# Copyright (C) 2018-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import re
import pytest

from swh.core.api import remote_api_endpoint, RPCClient

from .test_serializers import ExtraType, extra_encoders, extra_decoders


@pytest.fixture
def rpc_client(requests_mock):
    class TestStorage:
        @remote_api_endpoint('test_endpoint_url')
        def test_endpoint(self, test_data, db=None, cur=None):
            ...

        @remote_api_endpoint('path/to/endpoint')
        def something(self, data, db=None, cur=None):
            ...

        @remote_api_endpoint('serializer_test')
        def serializer_test(self, data, db=None, cur=None):
            ...

    class Testclient(RPCClient):
        backend_class = TestStorage
        extra_type_encoders = extra_encoders
        extra_type_decoders = extra_decoders

    def callback(request, context):
        assert request.headers['Content-Type'] == 'application/x-msgpack'
        context.headers['Content-Type'] = 'application/x-msgpack'
        if request.path == '/test_endpoint_url':
            context.content = b'\xa3egg'
        elif request.path == '/path/to/endpoint':
            context.content = b'\xa4spam'
        elif request.path == '/serializer_test':
            context.content = (
                b'\x82\xc4\x07swhtype\xa9extratype'
                b'\xc4\x01d\x92\x81\xa4spam\xa3egg\xa3qux')
        else:
            assert False
        return context.content

    requests_mock.post(re.compile('mock://example.com/'),
                       content=callback)

    return Testclient(url='mock://example.com')


def test_client(rpc_client):

    assert hasattr(rpc_client, 'test_endpoint')
    assert hasattr(rpc_client, 'something')

    res = rpc_client.test_endpoint('spam')
    assert res == 'egg'
    res = rpc_client.test_endpoint(test_data='spam')
    assert res == 'egg'

    res = rpc_client.something('whatever')
    assert res == 'spam'
    res = rpc_client.something(data='whatever')
    assert res == 'spam'


def test_client_extra_serializers(rpc_client):
    res = rpc_client.serializer_test(['foo', ExtraType('bar', b'baz')])
    assert res == ExtraType({'spam': 'egg'}, 'qux')



# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import json
from typing import Any, Callable, List, Tuple
import unittest
from uuid import UUID

import arrow
import requests
import requests_mock

from swh.core.api.serializers import (
    SWHJSONDecoder,
    SWHJSONEncoder,
    msgpack_dumps,
    msgpack_loads,
    decode_response
)


class ExtraType:
    def __init__(self, arg1, arg2):
        self.arg1 = arg1
        self.arg2 = arg2

    def __repr__(self):
        return f'ExtraType({self.arg1}, {self.arg2})'

    def __eq__(self, other):
        return isinstance(other, ExtraType) \
            and (self.arg1, self.arg2) == (other.arg1, other.arg2)


extra_encoders: List[Tuple[type, str, Callable[..., Any]]] = [
    (ExtraType, 'extratype', lambda o: (o.arg1, o.arg2))
]


extra_decoders = {
    'extratype': lambda o: ExtraType(*o),
}


class Serializers(unittest.TestCase):
    def setUp(self):
        self.tz = datetime.timezone(datetime.timedelta(minutes=118))

        self.data = {
            'bytes': b'123456789\x99\xaf\xff\x00\x12',
            'datetime_naive': datetime.datetime(2015, 1, 1, 12, 4, 42, 231455),
            'datetime_tz': datetime.datetime(2015, 3, 4, 18, 25, 13, 1234,
                                             tzinfo=self.tz),
            'datetime_utc': datetime.datetime(2015, 3, 4, 18, 25, 13, 1234,
                                              tzinfo=datetime.timezone.utc),
            'datetime_delta': datetime.timedelta(64),
            'arrow_date': arrow.get('2018-04-25T16:17:53.533672+00:00'),
            'swhtype': 'fake',
            'swh_dict': {'swhtype': 42, 'd': 'test'},
            'random_dict': {'swhtype': 43},
            'uuid': UUID('cdd8f804-9db6-40c3-93ab-5955d3836234'),
        }

        self.encoded_data = {
            'bytes': {'swhtype': 'bytes', 'd': 'F)}kWH8wXmIhn8j01^'},
            'datetime_naive': {'swhtype': 'datetime',
                               'd': '2015-01-01T12:04:42.231455'},
            'datetime_tz': {'swhtype': 'datetime',
                            'd': '2015-03-04T18:25:13.001234+01:58'},
            'datetime_utc': {'swhtype': 'datetime',
                             'd': '2015-03-04T18:25:13.001234+00:00'},
            'datetime_delta': {'swhtype': 'timedelta',
                               'd': {'days': 64, 'seconds': 0,
                                     'microseconds': 0}},
            'arrow_date': {'swhtype': 'arrow',
                           'd': '2018-04-25T16:17:53.533672+00:00'},
            'swhtype': 'fake',
            'swh_dict': {'swhtype': 42, 'd': 'test'},
            'random_dict': {'swhtype': 43},
            'uuid': {'swhtype': 'uuid',
                     'd': 'cdd8f804-9db6-40c3-93ab-5955d3836234'},
        }

        self.legacy_msgpack = {
            'bytes': b'\xc4\x0e123456789\x99\xaf\xff\x00\x12',
            'datetime_naive': (
                b'\x82\xc4\x0c__datetime__\xc3\xc4\x01s\xba'
                b'2015-01-01T12:04:42.231455'
            ),
            'datetime_tz': (
                b'\x82\xc4\x0c__datetime__\xc3\xc4\x01s\xd9 '
                b'2015-03-04T18:25:13.001234+01:58'
            ),
            'datetime_utc': (
                b'\x82\xc4\x0c__datetime__\xc3\xc4\x01s\xd9 '
                b'2015-03-04T18:25:13.001234+00:00'
            ),
            'datetime_delta': (
                b'\x82\xc4\r__timedelta__\xc3\xc4\x01s\x83\xa4'
                b'days@\xa7seconds\x00\xacmicroseconds\x00'
            ),
            'arrow_date': (
                b'\x82\xc4\t__arrow__\xc3\xc4\x01s\xd9 '
                b'2018-04-25T16:17:53.533672+00:00'
            ),
            'swhtype': b'\xa4fake',
            'swh_dict': b'\x82\xa7swhtype*\xa1d\xa4test',
            'random_dict': b'\x81\xa7swhtype+',
            'uuid': (
                b'\x82\xc4\x08__uuid__\xc3\xc4\x01s\xd9$'
                b'cdd8f804-9db6-40c3-93ab-5955d3836234'
            ),
        }

        self.generator = (i for i in range(5))
        self.gen_lst = list(range(5))

    def test_round_trip_json(self):
        data = json.dumps(self.data, cls=SWHJSONEncoder)
        self.assertEqual(self.data, json.loads(data, cls=SWHJSONDecoder))

    def test_round_trip_json_extra_types(self):
        original_data = [ExtraType('baz', self.data), 'qux']

        data = json.dumps(original_data, cls=SWHJSONEncoder,
                          extra_encoders=extra_encoders)
        self.assertEqual(
            original_data,
            json.loads(
                data, cls=SWHJSONDecoder, extra_decoders=extra_decoders))

    def test_encode_swh_json(self):
        data = json.dumps(self.data, cls=SWHJSONEncoder)
        self.assertEqual(self.encoded_data, json.loads(data))

    def test_round_trip_msgpack(self):
        original_data = {
            **self.data,
            'none_dict_key': {None: 42},
        }
        data = msgpack_dumps(original_data)
        self.assertEqual(original_data, msgpack_loads(data))

    def test_round_trip_msgpack_extra_types(self):
        original_data = [ExtraType('baz', self.data), 'qux']

        data = msgpack_dumps(original_data, extra_encoders=extra_encoders)
        self.assertEqual(
            original_data, msgpack_loads(data, extra_decoders=extra_decoders))

    def test_generator_json(self):
        data = json.dumps(self.generator, cls=SWHJSONEncoder)
        self.assertEqual(self.gen_lst, json.loads(data, cls=SWHJSONDecoder))

    def test_generator_msgpack(self):
        data = msgpack_dumps(self.generator)
        self.assertEqual(self.gen_lst, msgpack_loads(data))

    @requests_mock.Mocker()
    def test_decode_response_json(self, mock_requests):
        mock_requests.get('https://example.org/test/data',
                          json=self.encoded_data,
                          headers={'content-type': 'application/json'})
        response = requests.get('https://example.org/test/data')
        assert decode_response(response) == self.data

    def test_decode_legacy_msgpack(self):
        for k, v in self.legacy_msgpack.items():
            assert msgpack_loads(v) == self.data[k]



# Copyright (C) 2015-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import base64
import datetime
import json
import traceback
import types
from uuid import UUID

import arrow
import iso8601
import msgpack

from typing import Any, Dict, Union, Tuple
from requests import Response


ENCODERS = [
    (arrow.Arrow, 'arrow', arrow.Arrow.isoformat),
    (datetime.datetime, 'datetime', datetime.datetime.isoformat),
    (datetime.timedelta, 'timedelta', lambda o: {
        'days': o.days,
        'seconds': o.seconds,
        'microseconds': o.microseconds,
    }),
    (UUID, 'uuid', str),

    # Only for JSON:
    (bytes, 'bytes', lambda o: base64.b85encode(o).decode('ascii')),
]

DECODERS = {
    'arrow': arrow.get,
    'datetime': lambda d: iso8601.parse_date(d, default_timezone=None),
    'timedelta': lambda d: datetime.timedelta(**d),
    'uuid': UUID,

    # Only for JSON:
    'bytes': base64.b85decode,
}


def encode_data_client(data: Any, extra_encoders=None) -> bytes:
    try:
        return msgpack_dumps(data, extra_encoders=extra_encoders)
    except OverflowError as e:
        raise ValueError('Limits were reached. Please, check your input.\n' +
                         str(e))


def decode_response(response: Response, extra_decoders=None) -> Any:
    content_type = response.headers['content-type']

    if content_type.startswith('application/x-msgpack'):
        r = msgpack_loads(response.content,
                          extra_decoders=extra_decoders)
    elif content_type.startswith('application/json'):
        r = json_loads(response.text,
                       extra_decoders=extra_decoders)
    elif content_type.startswith('text/'):
        r = response.text
    else:
        raise ValueError('Wrong content type `%s` for API response'
                         % content_type)

    return r


class SWHJSONEncoder(json.JSONEncoder):
    """JSON encoder for data structures generated by Software Heritage.

    This JSON encoder extends the default Python JSON encoder and adds
    awareness for the following specific types:

    - bytes (get encoded as a Base85 string);
    - datetime.datetime (get encoded as an ISO8601 string).

    Non-standard types get encoded as a a dictionary with two keys:

    - swhtype with value 'bytes' or 'datetime';
    - d containing the encoded value.

    SWHJSONEncoder also encodes arbitrary iterables as a list
    (allowing serialization of generators).

    Caveats: Limitations in the JSONEncoder extension mechanism
    prevent us from "escaping" dictionaries that only contain the
    swhtype and d keys, and therefore arbitrary data structures can't
    be round-tripped through SWHJSONEncoder and SWHJSONDecoder.

    """

    def __init__(self, extra_encoders=None, **kwargs):
        super().__init__(**kwargs)
        self.encoders = ENCODERS
        if extra_encoders:
            self.encoders += extra_encoders

    def default(self, o: Any
                ) -> Union[Dict[str, Union[Dict[str, int], str]], list]:
        for (type_, type_name, encoder) in self.encoders:
            if isinstance(o, type_):
                return {
                    'swhtype': type_name,
                    'd': encoder(o),
                }
        try:
            return super().default(o)
        except TypeError as e:
            try:
                iterable = iter(o)
            except TypeError:
                raise e from None
            else:
                return list(iterable)


class SWHJSONDecoder(json.JSONDecoder):
    """JSON decoder for data structures encoded with SWHJSONEncoder.

    This JSON decoder extends the default Python JSON decoder,
    allowing the decoding of:

    - bytes (encoded as a Base85 string);
    - datetime.datetime (encoded as an ISO8601 string).

    Non-standard types must be encoded as a a dictionary with exactly
    two keys:

    - swhtype with value 'bytes' or 'datetime';
    - d containing the encoded value.

    To limit the impact our encoding, if the swhtype key doesn't
    contain a known value, the dictionary is decoded as-is.

    """

    def __init__(self, extra_decoders=None, **kwargs):
        super().__init__(**kwargs)
        self.decoders = DECODERS
        if extra_decoders:
            self.decoders = {**self.decoders, **extra_decoders}

    def decode_data(self, o: Any) -> Any:
        if isinstance(o, dict):
            if set(o.keys()) == {'d', 'swhtype'}:
                if o['swhtype'] == 'bytes':
                    return base64.b85decode(o['d'])
                decoder = self.decoders.get(o['swhtype'])
                if decoder:
                    return decoder(self.decode_data(o['d']))
            return {key: self.decode_data(value) for key, value in o.items()}
        if isinstance(o, list):
            return [self.decode_data(value) for value in o]
        else:
            return o

    def raw_decode(self, s: str, idx: int = 0) -> Tuple[Any, int]:
        data, index = super().raw_decode(s, idx)
        return self.decode_data(data), index


def json_dumps(data: Any, extra_encoders=None) -> str:
    return json.dumps(data, cls=SWHJSONEncoder,
                      extra_encoders=extra_encoders)


def json_loads(data: str, extra_decoders=None) -> Any:
    return json.loads(data, cls=SWHJSONDecoder,
                      extra_decoders=extra_decoders)


def msgpack_dumps(data: Any, extra_encoders=None) -> bytes:
    """Write data as a msgpack stream"""
    encoders = ENCODERS
    if extra_encoders:
        encoders += extra_encoders

    def encode_types(obj):
        if isinstance(obj, types.GeneratorType):
            return list(obj)

        for (type_, type_name, encoder) in encoders:
            if isinstance(obj, type_):
                return {
                    b'swhtype': type_name,
                    b'd': encoder(obj),
                }
        return obj

    return msgpack.packb(data, use_bin_type=True, default=encode_types)


def msgpack_loads(data: bytes, extra_decoders=None) -> Any:
    """Read data as a msgpack stream"""
    decoders = DECODERS
    if extra_decoders:
        decoders = {**decoders, **extra_decoders}

    def decode_types(obj):
        # Support for current encodings
        if set(obj.keys()) == {b'd', b'swhtype'}:
            decoder = decoders.get(obj[b'swhtype'])
            if decoder:
                return decoder(obj[b'd'])

        # Support for legacy encodings
        if b'__datetime__' in obj and obj[b'__datetime__']:
            return iso8601.parse_date(obj[b's'], default_timezone=None)
        if b'__uuid__' in obj and obj[b'__uuid__']:
            return UUID(obj[b's'])
        if b'__timedelta__' in obj and obj[b'__timedelta__']:
            return datetime.timedelta(**obj[b's'])
        if b'__arrow__' in obj and obj[b'__arrow__']:
            return arrow.get(obj[b's'])

        # Fallthrough
        return obj

    try:
        try:
            return msgpack.unpackb(data, raw=False,
                                   object_hook=decode_types,
                                   strict_map_key=False)
        except TypeError:  # msgpack < 0.6.0
            return msgpack.unpackb(data, raw=False,
                                   object_hook=decode_types)
    except TypeError:  # msgpack < 0.5.2
        return msgpack.unpackb(data, encoding='utf-8',
                               object_hook=decode_types)


def exception_to_dict(exception):
    tb = traceback.format_exception(None, exception, exception.__traceback__)
    return {
        'exception': {
            'type': type(exception).__name__,
            'args': exception.args,
            'message': str(exception),
            'traceback': tb,
        }
    }
